from uuid import uuid4
import altair as alt
from altair.utils.html import spec_to_html


class RowLimitError(ValueError):
    def __init__(self, row_limit):
        self.row_limit = row_limit

    def __str__(self):
        return (
            f"Limit of {self.row_limit} rows was exceeded.\n"
            "Either introduce an aggregation to reduce the number of rows sent to the client or\n"
            "increase the row_limit argument to the vegafusion.enable() function"
        )


def vegafusion_mime_renderer(spec, mimetype="html", row_limit=None, embed_options=None):
    return spec_to_mime_bundle(
        spec,
        mimetype=mimetype,
        row_limit=row_limit,
        embed_options=embed_options
    )


def spec_to_mime_bundle(
        spec,
        mimetype="html",
        row_limit=None,
        embed_options=None,
        html_template="universal",
        full_html=False,
        scale=1,
):
    from . import transformer, runtime, local_tz, vegalite_compilers, altair_vl_version
    vega_spec = vegalite_compilers.get()(spec)

    inline_datasets = transformer.get_inline_datasets_for_spec(vega_spec)
    tx_vega_spec, warnings = runtime.pre_transform_spec(
        vega_spec,
        local_tz.get_local_tz(),
        row_limit=row_limit,
        inline_datasets=inline_datasets
    )

    for warning in warnings:
        if warning.get("type", "") == "RowLimitExceeded":
            raise RowLimitError(row_limit)

    # Handle default embed options
    embed_options = embed_options or {}
    embed_options = dict({"mode": "vega"}, **embed_options)

    if mimetype == "vega":
        vega_mimetype = "application/vnd.vega.v5+json"
        return (
            {vega_mimetype: tx_vega_spec},
            {vega_mimetype: {"embed_options": embed_options}}
        )
    elif mimetype == "html":
        output_div = f"altair-viz-{uuid4().hex}"
        html = spec_to_html(
            tx_vega_spec,
            mode="vega",
            vega_version="5",
            vegalite_version=altair_vl_version(),
            vegaembed_version="6",
            fullhtml=full_html,
            output_div=output_div,
            template=html_template,
            embed_options=embed_options
        )
        return {"text/html": html}
    elif mimetype == "svg":
        import vl_convert as vlc
        svg = vlc.vega_to_svg(tx_vega_spec)
        return {"image/svg+xml": svg}
    elif mimetype == "png":
        import vl_convert as vlc
        png = vlc.vega_to_png(tx_vega_spec, scale=scale)
        return {"image/png": png}
    else:
        raise ValueError(f"Unsupported mimetype: {mimetype}")


alt.renderers.register('vegafusion-mime', vegafusion_mime_renderer)



import pandas as pd
import psutil
import pyarrow as pa
from typing import Union
from .connection import SqlConnection
from .dataset import SqlDataset, DataFrameDataset
from .datasource import PandasDatasource, DfiDatasource
from .evaluation import get_mark_group_for_scope
from .transformer import import_pyarrow_interchange, to_arrow_table
from .local_tz import get_local_tz

try:
    from duckdb import DuckDBPyConnection
except ImportError:
    DuckDBPyConnection = None

try:
    import polars as pl
except ImportError:
    pl = None

from typing import TypedDict, List, Literal, Any


def _all_datasets_have_type(inline_datasets, types):
    if not inline_datasets:
        # If there are no inline datasets, return false
        # (we want the default pandas behavior in this case)
        return False
    else:
        for dataset in inline_datasets.values():
            if not isinstance(dataset, types):
                return False
        return True


class VariableUpdate(TypedDict):
    name: str
    namespace: Literal["data", "signal"]
    scope: List[int]
    value: Any


class Watch(TypedDict):
    name: str
    namespace: Literal["data", "signal"]
    scope: List[int]


class WatchPlan(TypedDict):
    client_to_server: List[Watch]
    server_to_client: List[Watch]


class PreTransformWarning(TypedDict):
    type: str
    message: str


class ChartState:
    def __init__(self, chart_state):
        self._chart_state = chart_state

    def update(self, client_updates: List[VariableUpdate]) -> List[VariableUpdate]:
        """Update chart state with updates from the client

        :param client_updates: List of VariableUpdate values from the client
        :return: list of VariableUpdates that should be pushed to the client
        """
        return self._chart_state.update(client_updates)

    def get_watch_plan(self) -> WatchPlan:
        """Get ChartState's watch plan

        The watch plan specifies the signals and datasets that should be communicated
        between ChartState and client to preserve the input Vega spec's interactivity
        :return: WatchPlan
        """
        return self._chart_state.get_watch_plan()

    def get_transformed_spec(self) -> dict:
        """Get initial transformed spec

        Get the initial transformed spec. This is equivalent to the spec that would
        be produced by vf.runtime.pre_transform_spec()
        """
        return self._chart_state.get_transformed_spec()

    def get_warnings(self) -> List[PreTransformWarning]:
        """Get transformed spec warnings

        :return: A list of warnings as dictionaries. Each warning dict has a 'type'
           key indicating the warning type, and a 'message' key containing
           a description of the warning. Potential warning types include:
            'RowLimitExceeded': Some datasets in resulting Vega specification
                have been truncated to the provided row limit
            'BrokenInteractivity': Some interactive features may have been
                broken in the resulting Vega specification
            'Unsupported': No transforms in the provided Vega specification were
                eligible for pre-transforming
        """
        return self._chart_state.get_warnings()

    def get_server_spec(self) -> dict:
        """Get server spec"""
        return self._chart_state.get_server_spec()

    def get_client_spec(self) -> dict:
        """Get client spec"""
        return self._chart_state.get_client_spec()


class VegaFusionRuntime:
    def __init__(self, cache_capacity, memory_limit, worker_threads, connection=None):
        self._embedded_runtime = None
        self._grpc_channel = None
        self._grpc_query = None
        self._cache_capacity = cache_capacity
        self._memory_limit = memory_limit
        self._worker_threads = worker_threads
        self._connection = connection

    @property
    def embedded_runtime(self):
        if self._embedded_runtime is None:
            # Try to initialize an embedded runtime
            from vegafusion_embed import PyVegaFusionRuntime
            self._embedded_runtime = PyVegaFusionRuntime(
                self.cache_capacity, self.memory_limit, self.worker_threads, connection=self._connection
            )
        return self._embedded_runtime

    def set_connection(self, connection: Union[str, SqlConnection, DuckDBPyConnection] = "datafusion"):
        """
        Sets the connection to use to evaluate Vega data transformations.

        Named tables returned by the connection's `tables` method may be referenced in Vega/Altair
        chart specifications using special dataset URLs. For example, if the connection's `tables`
        method returns a dictionary that includes "tableA" as a key, then this table may be
        referenced in a chart specification using the URL "table://tableA" or
        "vegafusion+dataset://tableA".

        :param connection: One of:
          - An instance of vegafusion.connection.SqlConnection
          - An instance of a duckdb connection
          - A string, one of:
                - "datafusion" (default)
                - "duckdb"
        """
        if isinstance(connection, str):
            if connection == "datafusion":
                # Connection of None uses DataFusion
                connection = None
            elif connection == "duckdb":
                from vegafusion.connection.duckdb import DuckDbConnection
                connection = DuckDbConnection()
            else:
                raise ValueError(f"Unsupported connection name: {connection}")
        elif DuckDBPyConnection is not None and isinstance(connection, DuckDBPyConnection):
            from vegafusion.connection.duckdb import DuckDbConnection
            connection = DuckDbConnection(connection)
        elif not isinstance(connection, SqlConnection):
            raise ValueError(
                "connection argument must be a string or an instance of SqlConnection\n"
                f"Received value of type {type(connection).__name__}: {connection}"
            )

        self._connection = connection
        self.reset()

    def grpc_connect(self, channel):
        """
        Connect to a VegaFusion server over gRPC using the provided gRPC channel

        :param channel: grpc.Channel instance configured with the address of a running VegaFusion server
        """
        # TODO: check channel type
        self._grpc_channel = channel

    @property
    def using_grpc(self):
        return self._grpc_channel is not None

    @property
    def grpc_query(self):
        if self._grpc_channel is None:
            raise ValueError(
                "No grpc channel registered. Use runtime.grpc_connect to provide a grpc channel"
            )

        if self._grpc_query is None:
            self._grpc_query = self._grpc_channel.unary_unary(
                '/services.VegaFusionRuntime/TaskGraphQuery',
            )
        return self._grpc_query

    def process_request_bytes(self, request):
        if self._grpc_channel:
            return self.grpc_query(request)
        else:
            # No grpc channel, get or initialize an embedded runtime
            return self.embedded_runtime.process_request_bytes(request)

    def _import_or_register_inline_datasets(self, inline_datasets=None):
        inline_datasets = inline_datasets or dict()
        imported_inline_datasets = dict()
        for name, value in inline_datasets.items():
            if isinstance(value, SqlDataset):
                imported_inline_datasets[name] = value
            elif isinstance(value, DataFrameDataset):
                imported_inline_datasets[name] = value
            elif isinstance(value, pa.Table):
                if self._connection is not None:
                    try:
                        # Try registering Arrow Table if supported
                        self._connection.register_arrow(name, value, temporary=True)
                        continue
                    except ValueError:
                        pass

                imported_inline_datasets[name] = DfiDatasource(value)
            elif isinstance(value, pd.DataFrame):
                if self._connection is not None:
                    try:
                        # Try registering DataFrame if supported
                        self._connection.register_pandas(name, value, temporary=True)
                        continue
                    except ValueError:
                        pass

                imported_inline_datasets[name] = PandasDatasource(value)
            elif hasattr(value, "__dataframe__"):
                imported_inline_datasets[name] = DfiDatasource(value)
            else:
                raise ValueError(f"Unsupported DataFrame type: {type(value)}")

        return imported_inline_datasets

    def build_pre_transform_spec_plan(
            self,
            spec,
            preserve_interactivity=True,
            keep_signals=None,
            keep_datasets=None,
    ):
        """
        Diagnostic function that returns the plan used by the pre_transform_spec method

        :param spec: A Vega specification dict or JSON string
        :param preserve_interactivity: If True (default) then the interactive behavior of
            the chart will pre preserved. This requires that all the data that participates
            in interactions be included in the resulting spec rather than being pre-transformed.
            If False, then all possible data transformations are applied even if they break
            the original interactive behavior of the chart.
        :param keep_signals: Signals from the input spec that must be included in the
            pre-transformed spec. A list with elements that are either:
              - The name of a top-level signal as a string
              - A two-element tuple where the first element is the name of a signal as a string
                and the second element is the nested scope of the dataset as a list of integers
        :param keep_datasets: Datasets from the input spec that must be included in the
            pre-transformed spec. A list with elements that are either:
              - The name of a top-level dataset as a string
              - A two-element tuple where the first element is the name of a dataset as a string
                and the second element is the nested scope of the dataset as a list of integers
        :return:
            dict with keys:
                - "client_spec": Planned client spec
                - "server_spec: Planned server spec
                - "comm_plan": Communication plan
                - "warnings": List of planner warnings
        """
        if self._grpc_channel:
            raise ValueError("build_pre_transform_spec_plan not yet supported over gRPC")
        else:
            # Parse input keep signals and datasets
            keep_signals = parse_variables(keep_signals)
            keep_datasets = parse_variables(keep_datasets)
            return self.embedded_runtime.build_pre_transform_spec_plan(
                spec,
                preserve_interactivity=preserve_interactivity,
                keep_signals=keep_signals,
                keep_datasets=keep_datasets,
            )

    def pre_transform_spec(
        self,
        spec,
        local_tz=None,
        default_input_tz=None,
        row_limit=None,
        preserve_interactivity=True,
        inline_datasets=None,
        keep_signals=None,
        keep_datasets=None,
        data_encoding_threshold=None,
        data_encoding_format="pyarrow",
    ):
        """
        Evaluate supported transforms in an input Vega specification and produce a new
        specification with pre-transformed datasets included inline.

        :param spec: A Vega specification dict or JSON string
        :param local_tz: Name of timezone to be considered local. E.g. 'America/New_York'.
            Defaults to the value of vf.get_local_tz(), which defaults to the system timezone
            if one can be determined.
        :param default_input_tz: Name of timezone (e.g. 'America/New_York') that naive datetime
            strings should be interpreted in. Defaults to `local_tz`.
        :param row_limit: Maximum number of dataset rows to include in the returned
            specification. If exceeded, datasets will be truncated to this number of rows
            and a RowLimitExceeded warning will be included in the resulting warnings list
        :param preserve_interactivity: If True (default) then the interactive behavior of
            the chart will pre preserved. This requires that all the data that participates
            in interactions be included in the resulting spec rather than being pre-transformed.
            If False, then all possible data transformations are applied even if they break
            the original interactive behavior of the chart.
        :param inline_datasets: A dict from dataset names to pandas DataFrames or pyarrow
            Tables. Inline datasets may be referenced by the input specification using
            the following url syntax 'vegafusion+dataset://{dataset_name}' or
            'table://{dataset_name}'.
        :param keep_signals: Signals from the input spec that must be included in the
            pre-transformed spec. A list with elements that are either:
              - The name of a top-level signal as a string
              - A two-element tuple where the first element is the name of a signal as a string
                and the second element is the nested scope of the dataset as a list of integers
        :param keep_datasets: Datasets from the input spec that must be included in the
            pre-transformed spec. A list with elements that are either:
              - The name of a top-level dataset as a string
              - A two-element tuple where the first element is the name of a dataset as a string
                and the second element is the nested scope of the dataset as a list of integers
        :param data_encoding_threshold: threshold for encoding datasets
            When length of pre-transformed datasets exceeds data_encoding_threshold, datasets
            are encoded into an alternative format (as determined by the data_encoding_format
            argument). When None (the default), pre-transformed datasets are never encoded and
            are always included as JSON compatible lists of dictionaries.
        :param data_encoding_format: format of encoded datasets
            Format to use to encode datasets with length exceeding the data_encoding_threshold
            argument.
                - "pyarrow": Encode datasets as pyarrow Tables. Not JSON compatible.
                - "arrow-ipc": Encode datasets as bytes in Arrow IPC format. Not JSON compatible.
                - "arrow-ipc-base64": Encode datasets as strings in base64 encoded Arrow IPC format.
                    JSON compatible.
        :return:
            Two-element tuple:
                0. A string containing the JSON representation of a Vega specification
                   with pre-transformed datasets included inline
                1. A list of warnings as dictionaries. Each warning dict has a 'type'
                   key indicating the warning type, and a 'message' key containing
                   a description of the warning. Potential warning types include:
                    'RowLimitExceeded': Some datasets in resulting Vega specification
                        have been truncated to the provided row limit
                    'BrokenInteractivity': Some interactive features may have been
                        broken in the resulting Vega specification
                    'Unsupported': No transforms in the provided Vega specification were
                        eligible for pre-transforming
        """
        if self._grpc_channel:
            raise ValueError("pre_transform_spec not yet supported over gRPC")
        else:
            local_tz = local_tz or get_local_tz()
            imported_inline_dataset = self._import_or_register_inline_datasets(inline_datasets)

            # Parse input keep signals and datasets
            keep_signals = parse_variables(keep_signals)
            keep_datasets = parse_variables(keep_datasets)
            try:
                if data_encoding_threshold is None:
                    new_spec, warnings = self.embedded_runtime.pre_transform_spec(
                        spec,
                        local_tz=local_tz,
                        default_input_tz=default_input_tz,
                        row_limit=row_limit,
                        preserve_interactivity=preserve_interactivity,
                        inline_datasets=imported_inline_dataset,
                        keep_signals=keep_signals,
                        keep_datasets=keep_datasets,
                    )
                else:
                    # Use pre_transform_extract to extract large datasets
                    new_spec, datasets, warnings = self.embedded_runtime.pre_transform_extract(
                        spec,
                        local_tz=local_tz,
                        default_input_tz=default_input_tz,
                        preserve_interactivity=preserve_interactivity,
                        extract_threshold=data_encoding_threshold,
                        extracted_format=data_encoding_format,
                        inline_datasets=imported_inline_dataset,
                        keep_signals=keep_signals,
                        keep_datasets=keep_datasets,
                    )

                    # Insert encoded datasets back into spec
                    for (name, scope, tbl) in datasets:
                        group = get_mark_group_for_scope(new_spec, scope)
                        for data in group.get("data", []):
                            if data.get("name", None) == name:
                                data["values"] = tbl

            finally:
                # Clean up temporary tables
                if self._connection is not None:
                    self._connection.unregister_temporary_tables()

            return new_spec, warnings

    def new_chart_state(
        self, spec, local_tz=None, default_input_tz=None, row_limit=None, inline_datasets=None
    ) -> ChartState:
        """
        Construct new ChartState object

        :param spec: A Vega specification dict or JSON string
        :param local_tz: Name of timezone to be considered local. E.g. 'America/New_York'.
            Defaults to the value of vf.get_local_tz(), which defaults to the system timezone
            if one can be determined.
        :param default_input_tz: Name of timezone (e.g. 'America/New_York') that naive datetime
            strings should be interpreted in. Defaults to `local_tz`.
        :param row_limit: Maximum number of dataset rows to include in the returned
            datasets. If exceeded, datasets will be truncated to this number of rows
            and a RowLimitExceeded warning will be included in the ChartState's warnings list
        :param inline_datasets: A dict from dataset names to pandas DataFrames or pyarrow
            Tables. Inline datasets may be referenced by the input specification using
            the following url syntax 'vegafusion+dataset://{dataset_name}' or
            'table://{dataset_name}'.
        :return: ChartState
        """
        if self._grpc_channel:
            raise ValueError("new_chart_state not yet supported over gRPC")
        else:
            local_tz = local_tz or get_local_tz()
            inline_arrow_dataset = self._import_or_register_inline_datasets(inline_datasets)
            return ChartState(
                self.embedded_runtime.new_chart_state(spec, local_tz, default_input_tz, row_limit, inline_arrow_dataset)
            )

    def pre_transform_datasets(
        self,
        spec,
        datasets,
        local_tz=None,
        default_input_tz=None,
        row_limit=None,
        inline_datasets=None
    ):
        """
        Extract the fully evaluated form of the requested datasets from a Vega specification
        as pandas DataFrames.

        :param spec: A Vega specification dict or JSON string
        :param datasets: A list with elements that are either:
          - The name of a top-level dataset as a string
          - A two-element tuple where the first element is the name of a dataset as a string
            and the second element is the nested scope of the dataset as a list of integers
        :param local_tz: Name of timezone to be considered local. E.g. 'America/New_York'.
            Defaults to the value of vf.get_local_tz(), which defaults to the system timezone
            if one can be determined.
        :param default_input_tz: Name of timezone (e.g. 'America/New_York') that naive datetime
            strings should be interpreted in. Defaults to `local_tz`.
        :param row_limit: Maximum number of dataset rows to include in the returned
            datasets. If exceeded, datasets will be truncated to this number of rows
            and a RowLimitExceeded warning will be included in the resulting warnings list
        :param inline_datasets: A dict from dataset names to pandas DataFrames or pyarrow
            Tables. Inline datasets may be referenced by the input specification using
            the following url syntax 'vegafusion+dataset://{dataset_name}' or
            'table://{dataset_name}'.
        :return:
            Two-element tuple:
                0. List of pandas DataFrames corresponding to the input datasets list
                1. A list of warnings as dictionaries. Each warning dict has a 'type'
                   key indicating the warning type, and a 'message' key containing
                   a description of the warning.
        """
        if self._grpc_channel:
            raise ValueError("pre_transform_datasets not yet supported over gRPC")
        else:
            local_tz = local_tz or get_local_tz()

            # Build input variables
            pre_tx_vars = parse_variables(datasets)

            # Serialize inline datasets
            inline_arrow_dataset = self._import_or_register_inline_datasets(inline_datasets)
            try:
                values, warnings = self.embedded_runtime.pre_transform_datasets(
                    spec,
                    pre_tx_vars,
                    local_tz=local_tz,
                    default_input_tz=default_input_tz,
                    row_limit=row_limit,
                    inline_datasets=inline_arrow_dataset
                )
            finally:
                # Clean up registered tables (both inline and internal temporary tables)
                if self._connection is not None:
                    self._connection.unregister_temporary_tables()

            if pl is not None and _all_datasets_have_type(inline_datasets, (pl.DataFrame, pl.LazyFrame)):
                # Deserialize values to Polars tables
                datasets = [pl.from_arrow(value) for value in values]

                # Localize datetime columns to UTC
                processed_datasets = []
                for df in datasets:
                    for name, dtype in zip(df.columns, df.dtypes):
                        if dtype == pl.Datetime:
                            df = df.with_columns(df[name].dt.replace_time_zone("UTC").dt.convert_time_zone(local_tz))
                    processed_datasets.append(df)

                return processed_datasets, warnings
            elif _all_datasets_have_type(inline_datasets, pa.Table):
                return values, warnings
            else:
                # Deserialize values to pandas DataFrames
                datasets = [value.to_pandas() for value in values]

                # Localize datetime columns to UTC
                for df in datasets:
                    for name, dtype in df.dtypes.items():
                        if dtype.kind == "M":
                            df[name] = df[name].dt.tz_localize("UTC").dt.tz_convert(local_tz)

                return datasets, warnings

    def pre_transform_extract(
        self,
        spec,
        local_tz=None,
        default_input_tz=None,
        preserve_interactivity=True,
        extract_threshold=20,
        extracted_format="pyarrow",
        inline_datasets=None,
        keep_signals=None,
        keep_datasets=None,
    ):
        """
        Evaluate supported transforms in an input Vega specification and produce a new
        specification with small pre-transformed datasets (under 100 rows) included inline
        and larger inline datasets (100 rows or more) are extracted into pyarrow tables.

        :param spec: A Vega specification dict or JSON string
        :param local_tz: Name of timezone to be considered local. E.g. 'America/New_York'.
            Defaults to the value of vf.get_local_tz(), which defaults to the system timezone
            if one can be determined.
        :param default_input_tz: Name of timezone (e.g. 'America/New_York') that naive datetime
            strings should be interpreted in. Defaults to `local_tz`.
        :param preserve_interactivity: If True (default) then the interactive behavior of
            the chart will pre preserved. This requires that all the data that participates
            in interactions be included in the resulting spec rather than being pre-transformed.
            If False, then all possible data transformations are applied even if they break
            the original interactive behavior of the chart.
        :param extract_threshold: Datasets with length below extract_threshold will be
            inlined
        :param extracted_format: The format for the extracted datasets
            The format for extracted datasets:
                - "pyarrow": pyarrow.Table
                - "arrow-ipc": bytes in arrow IPC format
                - "arrow-ipc-base64": base64 encoded arrow IPC format
        :param inline_datasets: A dict from dataset names to pandas DataFrames or pyarrow
            Tables. Inline datasets may be referenced by the input specification using
            the following url syntax 'vegafusion+dataset://{dataset_name}' or
            'table://{dataset_name}'.
        :param keep_signals: Signals from the input spec that must be included in the
            pre-transformed spec. A list with elements that are either:
              - The name of a top-level signal as a string
              - A two-element tuple where the first element is the name of a signal as a string
                and the second element is the nested scope of the dataset as a list of integers
        :param keep_datasets: Datasets from the input spec that must be included in the
            pre-transformed spec. A list with elements that are either:
              - The name of a top-level dataset as a string
              - A two-element tuple where the first element is the name of a dataset as a string
                and the second element is the nested scope of the dataset as a list of integers
        :return:
            Three-element tuple:
                0. A dict containing the JSON representation of the pre-transformed Vega
                   specification without pre-transformed datasets included inline
                1. Extracted datasets as a list of three element tuples
                    0. dataset name
                    1. dataset scope
                    2. pyarrow Table
                2. A list of warnings as dictionaries. Each warning dict has a 'type'
                   key indicating the warning type, and a 'message' key containing
                   a description of the warning. Potential warning types include:
                    'Planner': Planner warning
        """
        if self._grpc_channel:
            raise ValueError("pre_transform_spec not yet supported over gRPC")
        else:
            local_tz = local_tz or get_local_tz()

            inline_arrow_dataset = self._import_or_register_inline_datasets(inline_datasets)
            try:
                new_spec, datasets, warnings = self.embedded_runtime.pre_transform_extract(
                    spec,
                    local_tz=local_tz,
                    default_input_tz=default_input_tz,
                    preserve_interactivity=preserve_interactivity,
                    extract_threshold=extract_threshold,
                    extracted_format=extracted_format,
                    inline_datasets=inline_arrow_dataset,
                    keep_signals=keep_signals,
                    keep_datasets=keep_datasets,
                )
            finally:
                # Clean up temporary tables
                if self._connection is not None:
                    self._connection.unregister_temporary_tables()

            return new_spec, datasets, warnings

    def patch_pre_transformed_spec(self, spec1, pre_transformed_spec1, spec2):
        """
        Attempt to patch a Vega spec was returned by the pre_transform_spec method without
        rerunning the pre_transform_spec logic. When possible, this can be significantly
        faster than rerunning the pre_transform_spec method.

        :param spec1: The input Vega spec to a prior call to pre_transform_spec
        :param pre_transformed_spec1: The prior result of passing spec1 to pre_transform_spec
        :param spec2: A Vega spec that is assumed to be a small delta compared to spec1

        :return: dict or None
            If the delta between spec1 and spec2 is in the portions of spec1 that were not
            modified by pre_transform_spec, then this delta can be applied cleanly to
            pre_transform_spec1 and the result is returned. If the delta cannot be
            applied cleanly, None is returned and spec2 should be passed through the
            pre_transform_spec method.
        """
        if self._grpc_channel:
            raise ValueError("patch_pre_transformed_spec not yet supported over gRPC")
        else:
            pre_transformed_spec2 = self.embedded_runtime.patch_pre_transformed_spec(
                spec1, pre_transformed_spec1, spec2
            )
            return pre_transformed_spec2

    @property
    def worker_threads(self):
        return self._worker_threads

    @worker_threads.setter
    def worker_threads(self, value):
        """
        Restart the runtime with the specified number of worker threads

        :param threads: Number of threads for the new runtime
        """
        if value != self._worker_threads:
            self._worker_threads = value
            self.reset()

    @property
    def total_memory(self):
        if self._embedded_runtime:
            return self._embedded_runtime.total_memory()
        else:
            return None

    @property
    def _protected_memory(self):
        if self._embedded_runtime:
            return self._embedded_runtime.protected_memory()
        else:
            return None

    @property
    def _probationary_memory(self):
        if self._embedded_runtime:
            return self._embedded_runtime.probationary_memory()
        else:
            return None

    @property
    def size(self):
        if self._embedded_runtime:
            return self._embedded_runtime.size()
        else:
            return None

    @property
    def memory_limit(self):
        return self._memory_limit

    @memory_limit.setter
    def memory_limit(self, value):
        """
        Restart the runtime with the specified memory limit

        :param threads: Max approximate memory usage of cache
        """
        if value != self._memory_limit:
            self._memory_limit = value
            self.reset()

    @property
    def cache_capacity(self):
        return self._cache_capacity

    @cache_capacity.setter
    def cache_capacity(self, value):
        """
        Restart the runtime with the specified cache capacity

        :param threads: Max task graph values to cache
        """
        if value != self._cache_capacity:
            self._cache_capacity = value
            self.reset()

    def reset(self):
        if self._embedded_runtime is not None:
            self._embedded_runtime.clear_cache()
            self._embedded_runtime = None

    def __repr__(self):
        if self._grpc_channel:
            return f"VegaFusionRuntime(channel={self._grpc_channel})"
        else:
            return (
                f"VegaFusionRuntime("
                f"cache_capacity={self.cache_capacity}, worker_threads={self.worker_threads}"
                f")"
            )


def parse_variables(variables):
    # Build input variables
    pre_tx_vars = []
    if variables is None:
        return []

    if isinstance(variables, str):
        variables = [variables]

    err_msg = "Elements of variables argument must be strings are two-element tuples"
    for var in variables:
        if isinstance(var, str):
            pre_tx_vars.append((var, []))
        elif isinstance(var, (list, tuple)):
            if len(var) == 2:
                pre_tx_vars.append(tuple(var))
            else:
                raise ValueError(err_msg)
        else:
            raise ValueError(err_msg)
    return pre_tx_vars


runtime = VegaFusionRuntime(64, psutil.virtual_memory().total // 2, psutil.cpu_count())



import logging
import pyarrow as pa
from .sql import SqlDataset
from snowflake.snowpark import Table as SnowparkTable
from snowflake.snowpark.types import DataType as SnowparkDataType
from typing import Dict

from ..transformer import to_arrow_table

SNOWPARK_TO_PYARROW_TYPES: Dict[SnowparkDataType, pa.DataType] = {}


def get_snowpark_to_pyarrow_types():
    if not SNOWPARK_TO_PYARROW_TYPES:
        import snowflake.snowpark.types as sp_types

        SNOWPARK_TO_PYARROW_TYPES.update(
            {
                sp_types.LongType: pa.int64(),
                sp_types.BinaryType: pa.binary(),
                sp_types.BooleanType: pa.bool_(),
                sp_types.ByteType: pa.int8(),
                sp_types.StringType: pa.string(),
                sp_types.DateType: pa.date32(),
                sp_types.DoubleType: pa.float64(),
                sp_types.FloatType: pa.float32(),
                sp_types.IntegerType: pa.int32(),
                sp_types.ShortType: pa.int16(),
                sp_types.TimestampType: pa.timestamp("ms"),
            }
        )
    return SNOWPARK_TO_PYARROW_TYPES


def snowflake_field_to_pyarrow_type(provided_type: SnowparkDataType) -> pa.DataType:
    """
    Converts Snowflake types to PyArrow equivalent types, raising a ValueError if they aren't comparable.
    See https://docs.snowflake.com/en/sql-reference/intro-summary-data-types
    """
    from snowflake.snowpark.types import DecimalType as SnowparkDecimalType

    type_map = get_snowpark_to_pyarrow_types()
    if provided_type.__class__ in type_map:
        return type_map[provided_type.__class__]

    if isinstance(provided_type, SnowparkDecimalType):
        return pa.decimal128(provided_type.precision, provided_type.scale)
    else:
        raise ValueError(f"Unsupported Snowpark type: {provided_type}")


def snowpark_table_to_pyarrow_schema(table: SnowparkTable) -> pa.Schema:
    schema_fields = {}
    for name, field in zip(table.schema.names, table.schema.fields):
        normalised_name = name.strip('"')
        schema_fields[normalised_name] = snowflake_field_to_pyarrow_type(field.datatype)
    return pa.schema(schema_fields)


class SnowparkDataset(SqlDataset):
    def dialect(self) -> str:
        return "snowflake"

    def __init__(
        self, table: SnowparkTable, fallback: bool = True, verbose: bool = False
    ):
        if not isinstance(table, SnowparkTable):
            raise ValueError(
                f"SnowparkDataset accepts a snowpark Table. Received: {type(table)}"
            )
        self._table = table
        self._session = table._session

        self._fallback = fallback
        self._verbose = verbose
        self._table_name = table.table_name
        self._table_schema = snowpark_table_to_pyarrow_schema(self._table)

        self.logger = logging.getLogger("SnowparkDataset")

    def table_name(self) -> str:
        return self._table_name

    def table_schema(self) -> pa.Schema:
        return self._table_schema

    def fetch_query(self, query: str, schema: pa.Schema) -> pa.Table:
        self.logger.info(f"Snowflake Query:\n{query}\n")
        if self._verbose:
            print(f"Snowflake Query:\n{query}\n")

        sp_df = self._session.sql(query)
        batches = []
        for pd_batch in sp_df.to_pandas_batches():
            pa_tbl = to_arrow_table(pd_batch).cast(schema, safe=False)
            batches.extend(pa_tbl.to_batches())

        return pa.Table.from_batches(batches, schema)

    def fallback(self) -> bool:
        return self._fallback



import logging
import pyarrow as pa
from duckdb import DuckDBPyRelation
from .sql import SqlDataset
from ..connection.duckdb import duckdb_relation_to_schema


class DuckDbDataset(SqlDataset):
    def dialect(self) -> str:
        return "duckdb"

    def __init__(self, relation: DuckDBPyRelation, fallback: bool = True, verbose: bool = False):
        self._relation = relation
        self._fallback = fallback
        self._verbose = verbose
        self._table_name = "tbl"
        self._table_schema = duckdb_relation_to_schema(self._relation)
        self.logger = logging.getLogger("DuckDbDataset")

    def table_name(self) -> str:
        return self._table_name

    def table_schema(self) -> pa.Schema:
        return self._table_schema

    def fetch_query(self, query: str, schema: pa.Schema) -> pa.Table:
        self.logger.info(f"DuckDB Query:\n{query}\n")
        if self._verbose:
            print(f"DuckDB Query:\n{query}\n")

        return self._relation.query(self._table_name, query).to_arrow_table()

    def fallback(self) -> bool:
        return self._fallback



# https://altair-viz.github.io/gallery/stacked_bar_chart_sorted_segments.html

import altair as alt
from vega_datasets import data

source = data.barley()

alt.Chart(source).mark_bar().encode(
    x='sum(yield)',
    y='variety',
    color='site',
    order=alt.Order(
        # Sort the segments of the bars by this field
        'site',
        sort='ascending'
    )
)



import altair as alt
from vega_datasets import data

source = data.cars()

brush = alt.selection_interval()

points = alt.Chart(source).mark_point().encode(
    x='Horsepower:Q',
    y='Miles_per_Gallon:Q',
    color=alt.condition(brush, 'Origin:N', alt.value('lightgray'))
).add_params(
    brush
)

bars = alt.Chart(source).mark_bar().encode(
    y='Origin:N',
    color='Origin:N',
    x='count(Origin):Q'
).transform_filter(
    brush
)

points & bars


import altair as alt
from vega_datasets import data

source = data.movies.url

pts = alt.selection_point(encodings=['x'])

rect = alt.Chart(data.movies.url).mark_rect().encode(
    alt.X('IMDB_Rating:Q', bin=True),
    alt.Y('Rotten_Tomatoes_Rating:Q', bin=True),
    alt.Color('count()',
              scale=alt.Scale(scheme='greenblue'),
              legend=alt.Legend(title='Total Records')
              )
)

circ = rect.mark_point().encode(
    alt.ColorValue('grey'),
    alt.Size('count()',
             legend=alt.Legend(title='Records in Selection')
             )
).transform_filter(
    pts
)

bar = alt.Chart(source).mark_bar().encode(
    x='Major_Genre:N',
    y='count()',
    color=alt.condition(pts, alt.ColorValue("steelblue"), alt.ColorValue("grey"))
).properties(
    width=550,
    height=200
).add_params(pts)

alt.vconcat(
    rect + circ,
    bar
).resolve_legend(
    color="independent",
    size="independent"
)


# https://altair-viz.github.io/gallery/bar_rounded.html

import altair as alt
from vega_datasets import data

source = data.seattle_weather()

alt.Chart(source).mark_bar(
    cornerRadiusTopLeft=3,
    cornerRadiusTopRight=3
).encode(
    x='month(date):O',
    y='count():Q',
    color='weather:N'
)



# https://altair-viz.github.io/gallery/pacman_chart.html

import numpy as np
import altair as alt

alt.Chart().mark_arc(color="gold").encode(
    theta=alt.datum((5 / 8) * np.pi, scale=None),
    theta2=alt.datum((19 / 8) * np.pi),
    radius=alt.datum(100, scale=None),
)



# https://altair-viz.github.io/gallery/line_chart_with_generator.html

import altair as alt

source = alt.sequence(start=0, stop=12.71, step=0.1, as_='x')

alt.Chart(source).mark_line().transform_calculate(
    sin='sin(datum.x)',
    cos='cos(datum.x)'
).transform_fold(
    ['sin', 'cos']
).encode(
    x='x:Q',
    y='value:Q',
    color='key:N'
)



# https://altair-viz.github.io/gallery/beckers_barley_trellis_plot.html

import altair as alt
from vega_datasets import data

source = data.barley()

alt.Chart(source, title="The Morris Mistake").mark_point().encode(
    alt.X(
        'yield:Q',
        title="Barley Yield (bushels/acre)",
        scale=alt.Scale(zero=False),
        axis=alt.Axis(grid=False)
    ),
    alt.Y(
        'variety:N',
        title="",
        sort='-x',
        axis=alt.Axis(grid=True)
    ),
    color=alt.Color('year:N', legend=alt.Legend(title="Year")),
    facet=alt.Facet(
        'site:N',
        columns=2,
        title="",
        sort=alt.EncodingSortField(field='yield', op='sum', order='descending')
    ),
).properties(
    width=200,
    height=100,
).configure_view(stroke="transparent")



# https://altair-viz.github.io/gallery/cumulative_count_chart.html

import altair as alt
from vega_datasets import data

source = data.movies.url

alt.Chart(source).transform_window(
    cumulative_count="count()",
    sort=[{"field": "IMDB_Rating"}],
).mark_area().encode(
    x="IMDB_Rating:Q",
    y="cumulative_count:Q"
)



# https://altair-viz.github.io/gallery/histogram_responsive.html

import altair as alt
from vega_datasets import data

source = data.flights_5k.url

brush = alt.selection_interval(encodings=['x'])

base = alt.Chart(source).transform_calculate(
    time="hours(datum.date) + minutes(datum.date) / 60"
).mark_bar().encode(
    y='count():Q'
).properties(
    width=600,
    height=100
)

alt.vconcat(
    base.encode(
        alt.X('time:Q',
              bin=alt.Bin(maxbins=30, extent=brush),
              scale=alt.Scale(domain=brush)
              )
    ),
    base.encode(
        alt.X('time:Q', bin=alt.Bin(maxbins=30)),
    ).add_selection(brush)
)



# https://altair-viz.github.io/gallery/top_k_letters.html
# Added sort of initial data frame for consistent order of tie breakers

import altair as alt
import pandas as pd
import numpy as np

# Excerpt from A Tale of Two Cities; public domain text
text = """
It was the best of times, it was the worst of times, it was the age of wisdom,
it was the age of foolishness, it was the epoch of belief, it was the epoch of
incredulity, it was the season of Light, it was the season of Darkness, it was
the spring of hope, it was the winter of despair, we had everything before us,
we had nothing before us, we were all going direct to Heaven, we were all going
direct the other way - in short, the period was so far like the present period,
that some of its noisiest authorities insisted on its being received, for good
or for evil, in the superlative degree of comparison only.
"""

source = pd.DataFrame(
    {'letters': np.array([c for c in text if c.isalpha()])}
)

alt.Chart(source).transform_aggregate(
    count='count()',
    groupby=['letters']
).transform_window(
    rank='rank(count)',
    sort=[
        alt.SortField('count', order='descending'),
    ]
).transform_filter(
    alt.datum.rank < 10
).mark_bar().encode(
    y=alt.Y('letters:N', sort='-x'),
    x='count:Q',
)



# https://altair-viz.github.io/gallery/gantt_chart.html

import altair as alt
import pandas as pd

source = pd.DataFrame([
    {"task": "A", "start": 1, "end": 3},
    {"task": "B", "start": 3, "end": 8},
    {"task": "C", "start": 8, "end": 10}
])

alt.Chart(source).mark_bar().encode(
    x='start',
    x2='end',
    y='task'
)



# https://altair-viz.github.io/gallery/us_population_pyramid_over_time.html
import altair as alt
from vega_datasets import data

source = data.population.url

slider = alt.binding_range(min=1850, max=2000, step=10)
select_year = alt.selection_point(name='year', fields=['year'],
                                  bind=slider, value={'year': 2000})

base = alt.Chart(source).add_params(
    select_year
).transform_filter(
    select_year
).transform_calculate(
    gender=alt.expr.if_(alt.datum.sex == 1, 'Male', 'Female')
).properties(
    width=250
)


color_scale = alt.Scale(domain=['Male', 'Female'],
                        range=['#1f77b4', '#e377c2'])

left = base.transform_filter(
    alt.datum.gender == 'Female'
).encode(
    y=alt.Y('age:O', axis=None),
    x=alt.X('sum(people):Q',
            title='population',
            sort=alt.SortOrder('descending')),
    color=alt.Color('gender:N', scale=color_scale, legend=None)
).mark_bar().properties(title='Female')

middle = base.encode(
    y=alt.Y('age:O', axis=None),
    text=alt.Text('age:Q'),
).mark_text().properties(width=20)

right = base.transform_filter(
    alt.datum.gender == 'Male'
).encode(
    y=alt.Y('age:O', axis=None),
    x=alt.X('sum(people):Q', title='population'),
    color=alt.Color('gender:N', scale=color_scale, legend=None)
).mark_bar().properties(title='Male')

alt.concat(left, middle, right, spacing=5)


# https://altair-viz.github.io/gallery/connected_scatterplot.html

import altair as alt
from vega_datasets import data

source = data.driving()

alt.Chart(source).mark_line(point=True).encode(
    alt.X('miles', scale=alt.Scale(zero=False)),
    alt.Y('gas', scale=alt.Scale(zero=False)),
    order='year'
)



# https://altair-viz.github.io/gallery/select_detail.html

import altair as alt
import pandas as pd
import numpy as np

np.random.seed(0)

n_objects = 20
n_times = 50

# Create one (x, y) pair of metadata per object
locations = pd.DataFrame({
    'id': range(n_objects),
    'x': np.random.randn(n_objects),
    'y': np.random.randn(n_objects)
})

# Create a 50-element time-series for each object
timeseries = pd.DataFrame(np.random.randn(n_times, n_objects).cumsum(0),
                          columns=locations['id'],
                          index=pd.RangeIndex(0, n_times, name='time'))

# Melt the wide-form timeseries into a long-form view
timeseries = timeseries.reset_index().melt('time')

# Merge the (x, y) metadata into the long-form view
timeseries['id'] = timeseries['id'].astype(int)  # make merge not complain
data = pd.merge(timeseries, locations, on='id')

# Data is prepared, now make a chart

selector = alt.selection_single(empty='all', fields=['id'])

base = alt.Chart(data).properties(
    width=200,
    height=200
).add_selection(selector)

points = base.mark_point(filled=True, size=200).encode(
    x='mean(x)',
    y='mean(y)',
    color=alt.condition(selector, 'id:O', alt.value('lightgray'), legend=None),
)

timeseries = base.mark_line().encode(
    x='time',
    y=alt.Y('value', scale=alt.Scale(domain=(-15, 15))),
    color=alt.Color('id:O', legend=None)
).transform_filter(
    selector
)

points | timeseries


# https://altair-viz.github.io/gallery/gapminder_bubble_plot.html

import altair as alt
from vega_datasets import data

source = "https://raw.githubusercontent.com/vega/vega-datasets/main/data/gapminder-health-income.csv"

alt.Chart(source).mark_circle().encode(
    alt.X('income:Q', scale=alt.Scale(type='log')),
    alt.Y('health:Q', scale=alt.Scale(zero=False)),
    size='population:Q'
)



# https://altair-viz.github.io/gallery/airports.html

import altair as alt
from vega_datasets import data

airports = data.airports()
states = alt.topo_feature(data.us_10m.url, feature='states')

# US states background
background = alt.Chart(states).mark_geoshape(
    fill='lightgray',
    stroke='white'
).properties(
    width=500,
    height=300
).project('albersUsa')

# airport positions on background
points = alt.Chart(airports).mark_circle(
    size=10,
    color='steelblue'
).encode(
    longitude='longitude:Q',
    latitude='latitude:Q',
    tooltip=['name', 'city', 'state']
)

background + points



import pytest
import vegafusion as vf
import altair as alt
from altair.utils.execeval import eval_block
from pathlib import Path
here = Path(__file__).parent
altair_mocks_dir = here / "altair_mocks"


def test_row_limit():
    mock_path = altair_mocks_dir / "scatter" / "bubble_plot" / "mock.py"
    mock_src = mock_path.read_text("utf8")
    chart = eval_block(mock_src)

    # Dataset has 406 rows (before filtering out nulls) and this chart is not aggregated.
    # Limit of 500 rows should be fine
    with vf.enable(row_limit=500):
        chart._repr_mimebundle_()

    # Limit of 300 rows should raise RowLimitError
    with vf.enable(row_limit=300):
        with pytest.raises(vf.RowLimitError):
            chart._repr_mimebundle_()

    # Adding an aggregation should allow row limit to be much lower
    chart = chart.encode(
        alt.X("Horsepower", bin=True),
        alt.Y("Miles_per_Gallon", bin=True),
        alt.Size("count()")
    )
    with vf.enable(row_limit=50):
        chart._repr_mimebundle_()



# https://altair-viz.github.io/gallery/horizontal_stacked_bar_chart.html

import altair as alt
from vega_datasets import data

source = data.barley()

alt.Chart(source).mark_bar().encode(
    x='sum(yield)',
    y='variety',
    color='site'
)



# https://altair-viz.github.io/gallery/trellis_area_sort_array.html

import altair as alt
from vega_datasets import data

source = data.stocks()

alt.Chart(source).transform_filter(
    alt.datum.symbol != 'GOOG'
).mark_area().encode(
    x='date:T',
    y='price:Q',
    color='symbol:N',
    row=alt.Row('symbol:N', sort=['MSFT', 'AAPL', 'IBM', 'AMZN'])
).properties(height=50, width=400)



import json
import pathlib
from .renderer import spec_to_mime_bundle


def save_html(
    chart,
    file,
    embed_options=None,
    inline=False,
    full_html=True,
):
    """
    Save an Altair Chart to an HTML file after pre-applying data transformations
    and removing unused columns

    :param chart: alt.Chart
        The Altair chart to save
    :param file: str, pathlib.Path, or file-like object
        The file path to write the HTML file to, or a file-like object to write to
    :param embed_options: dict (default None)
        Dictionary of options to pass to vega-embed
    :param inline: boolean (default False)
        If False (default), the required JavaScript libraries are loaded from a
        CDN location. This results in a smaller file, but an internet connection
        is required to view the resulting HTML file.

        If True, all required JavaScript libraries are inlined.
        This results in a larger file, but no internet connection is required to view.
        inline=True requires Altair 5 and the altair_viewer package
    :param full_html: boolean (default True)
        If True, then a full html page is written. If False, then
        an HTML snippet that can be embedded into an HTML page is written.
    """
    from . import enable

    with enable():
        vegalite_spec = chart.to_dict()

    bundle = spec_to_mime_bundle(
        spec=vegalite_spec,
        mimetype="html",
        embed_options=embed_options,
        html_template="inline" if inline else "standard",
        full_html=full_html,
    )

    html = bundle["text/html"]
    write_file_or_filename(file, html)


def save_vega(
    chart,
    file,
    pretty=True,
):
    """
    Save an Altair Chart to a Vega JSON file after pre-applying data transformations
    and removing unused columns

    :param chart: alt.Chart
        The Altair chart to save
    :param file: str, pathlib.Path, or file-like object
        The file path to write the Vega JSON file to, or a file-like object to write to
    :param pretty: boolean (default True)
        If True, pretty-print the resulting JSON file. If False, write the smallest file possible
    """
    from . import enable

    with enable():
        vegalite_spec = chart.to_dict()

    (bundle, _) = spec_to_mime_bundle(
        spec=vegalite_spec,
        mimetype="vega",
    )

    vega_dict = bundle["application/vnd.vega.v5+json"]
    if pretty:
        vega_str = json.dumps(vega_dict, indent=2)
    else:
        vega_str = json.dumps(vega_dict)

    write_file_or_filename(file, vega_str)


def save_png(
    chart,
    file,
    scale=1,
):
    """
    Save an Altair Chart to a static PNG image file after pre-applying data transformations
    and removing unused columns

    :param chart: alt.Chart
        The Altair chart to save
    :param file: str, pathlib.Path, or file-like object
        The file path to write the PNG file to, or a file-like object to write to
    :param scale: float (default 1)
        Image scale factor
    """
    from . import enable

    with enable():
        vegalite_spec = chart.to_dict()

    bundle = spec_to_mime_bundle(
        spec=vegalite_spec,
        mimetype="png",
        scale=scale,
    )

    png = bundle["image/png"]
    write_file_or_filename(file, png, "wb")


def save_svg(
    chart,
    file,
):
    """
    Save an Altair Chart to a static SVG image file after pre-applying data transformations
    and removing unused columns

    :param chart: alt.Chart
        The Altair chart to save
    :param file: str, pathlib.Path, or file-like object
        The file path to write the SVG file to, or a file-like object to write to
    """
    from . import enable

    with enable():
        vegalite_spec = chart.to_dict()

    bundle = spec_to_mime_bundle(
        spec=vegalite_spec,
        mimetype="svg",
    )

    svg = bundle["image/svg+xml"]
    write_file_or_filename(file, svg, "w")


def write_file_or_filename(file, content, mode="w"):
    """Write content to file, whether file is a string, a pathlib Path or a
    file-like object"""
    if isinstance(file, str) or isinstance(file, pathlib.PurePath):
        with open(file, mode) as f:
            f.write(content)
    else:
        file.write(content)



from .runtime import runtime
from .transformer import to_feather, get_inline_datasets_for_spec
from .renderer import RowLimitError
from .local_tz import set_local_tz, get_local_tz
from .evaluation import transformed_data
from .save import save_html, save_vega, save_png, save_svg
from . import renderer
from .compilers import vegalite_compilers
import altair as alt

from ._version import __version__
from .utils import RendererTransformerEnabler

# Import optional subpackages
try:
    import vegafusion.jupyter
except ImportError:
    pass

try:
    import vegafusion.embed
except ImportError:
    pass


def altair_vl_version(vl_convert=False):
    """
    Get Altair's preferred Vega-Lite version

    :param vl_convert: If True, return a version string compatible with vl_convert
        (e.g. v4_17 rather than 4.17.0)
    :return: str with Vega-Lite version
    """
    from altair import SCHEMA_VERSION

    if vl_convert:
        # Compute VlConvert's vl_version string (of the form 'v5_2')
        # from SCHEMA_VERSION (of the form 'v5.2.0')
        return "_".join(SCHEMA_VERSION.split(".")[:2])
    else:
        # Return full version without leading v
        return SCHEMA_VERSION.lstrip("v")


def enable(mimetype="html", row_limit=10000, embed_options=None):
    """
    Enable the VegaFusion mime renderer so that all Altair charts
    are displayed using VegaFusion.

    :param mimetype: Mime type. One of:
        - "html" (default)
        - "vega"
        - "svg"
        - "png": Note, the PNG renderer can be quite slow for charts with lots of marks
    :param row_limit: Maximum number of rows (after transforms are applied) that may be
        included in the Vega specifications that will be displayed. An error will
        be raised if the limit is exceeded. None for unlimited.
    :param embed_options: dict (optional)
        Dictionary of options to pass to the vega-embed. Default
        entry is {'mode': 'vega'}.
    """
    return RendererTransformerEnabler(
        renderer_ctx=alt.renderers.enable(
            'vegafusion-mime', mimetype=mimetype, row_limit=row_limit, embed_options=embed_options
        ),
        data_transformer_ctx=alt.data_transformers.enable('vegafusion-inline'),
        repr_str=(
            "vegafusion.enable("
            f"mimetype={repr(mimetype)}, row_limit={row_limit}, embed_options={repr(embed_options)}"
            ")"
        )
    )


def enable_widget(
    debounce_wait=30,
    debounce_max_wait=60,
    data_dir="_vegafusion_data",
    download_source_link=None,
):
    """
    Enable the VegaFusion widget renderer so that all charts are displayed
    using VegaFusion.

    This isn't necessary in order to use the VegaFusionWidget directly

    :param debounce_wait: Debounce wait period for interactions (milliseconds)
    :param debounce_max_wait: Max debounce wait for interactions (milliseconds)
    :param download_source_link: URL of notebooks source code for inclusion
        in dropdown menu
    :param data_dir: Path that temporary feather files will be written to
    """
    from vegafusion.jupyter import enable
    return enable(
        download_source_link=download_source_link,
        debounce_wait=debounce_wait,
        debounce_max_wait=debounce_max_wait,
        data_dir=data_dir
    )


def disable():
    """
    Disable the VegaFusion data transformers and renderers so that Charts
    are not displayed using VegaFusion

    Equivalent to

    ```python
    import altair as alt
    alt.renderers.enable('default')
    alt.data_transformers.enable('default')
    ```

    This does not affect the behavior of VegaFusionWidget
    """
    return RendererTransformerEnabler(
        renderer_ctx=alt.renderers.enable('default'),
        data_transformer_ctx=alt.data_transformers.enable('default'),
        repr_str="vegafusion.disable()"
    )



# https://altair-viz.github.io/gallery/interactive_legend.html

import altair as alt
from vega_datasets import data

source = data.unemployment_across_industries.url

selection = alt.selection_multi(fields=['series'], bind='legend')

alt.Chart(source).mark_area().encode(
    alt.X('yearmonth(date):T', axis=alt.Axis(domain=False, format='%Y', tickSize=0)),
    alt.Y('sum(count):Q', stack='center', axis=None),
    alt.Color('series:N', scale=alt.Scale(scheme='category20b')),
    opacity=alt.condition(selection, alt.value(1), alt.value(0.2))
).add_selection(
    selection
)



# https://altair-viz.github.io/gallery/area_chart_gradient.html

import altair as alt
from vega_datasets import data

source = data.stocks()

alt.Chart(source).transform_filter(
    'datum.symbol==="GOOG"'
).mark_area(
    line={'color':'darkgreen'},
    color=alt.Gradient(
        gradient='linear',
        stops=[alt.GradientStop(color='white', offset=0),
               alt.GradientStop(color='darkgreen', offset=1)],
        x1=1,
        x2=1,
        y1=1,
        y2=0
    )
).encode(
    alt.X('date:T'),
    alt.Y('price:Q')
)



from typing import Iterable
import re
import pyarrow as pa
from pyarrow.interchange import from_dataframe

from typing import Any, Dict
from ._dfi_types import DtypeKind, DataFrame as DfiDataFrame
from .datasource import Datasource

# Taken from private pyarrow utilities
_PYARROW_DTYPES: Dict[DtypeKind, Dict[int, Any]] = {
    DtypeKind.INT: {8: pa.int8(),
                    16: pa.int16(),
                    32: pa.int32(),
                    64: pa.int64()},
    DtypeKind.UINT: {8: pa.uint8(),
                     16: pa.uint16(),
                     32: pa.uint32(),
                     64: pa.uint64()},
    DtypeKind.FLOAT: {16: pa.float16(),
                      32: pa.float32(),
                      64: pa.float64()},
    DtypeKind.BOOL: {1: pa.bool_(),
                     8: pa.uint8()},
    DtypeKind.STRING: {8: pa.string()},
}


def parse_datetime_format_str(format_str):
    """Parse datetime `format_str` to interpret the `data`."""

    # timestamp 'ts{unit}:tz'
    timestamp_meta = re.match(r"ts([smun]):(.*)", format_str)
    if timestamp_meta:
        unit, tz = timestamp_meta.group(1), timestamp_meta.group(2)
        if unit != "s":
            # the format string describes only a first letter of the unit, so
            # add one extra letter to convert the unit to numpy-style:
            # 'm' -> 'ms', 'u' -> 'us', 'n' -> 'ns'
            unit += "s"

        return unit, tz

    raise NotImplementedError(f"DateTime kind is not supported: {format_str}")


def map_date_type(data_type):
    """Map column date type to pyarrow date type. """
    kind, bit_width, f_string, _ = data_type

    if kind == DtypeKind.DATETIME:
        unit, tz = parse_datetime_format_str(f_string)
        return pa.timestamp(unit, tz=tz)
    else:
        pa_dtype = _PYARROW_DTYPES.get(kind, {}).get(bit_width, None)

        # Error if dtype is not supported
        if pa_dtype:
            return pa_dtype
        else:
            raise NotImplementedError(
                f"Conversion for {data_type} is not yet supported.")


class DfiDatasource(Datasource):
    def __init__(self, dataframe: DfiDataFrame):
        if hasattr(dataframe, "__dataframe__"):
            dataframe = dataframe.__dataframe__()
        fields = []
        for col in dataframe.column_names():
            field = dataframe.get_column_by_name(col)
            pa_type = map_date_type(field.dtype)
            fields.append(pa.field(col, pa_type))

        self._dataframe = dataframe
        self._schema = pa.schema(fields)

    def schema(self) -> pa.Schema:
        return self._schema

    def fetch(self, columns: Iterable[str]) -> pa.Table:
        columns = list(columns)
        projected_schema = pa.schema([f for f in self._schema if f.name in columns])
        table = from_dataframe(self._dataframe.select_columns_by_name(columns))
        return table.cast(projected_schema, safe=False)



# https://altair-viz.github.io/gallery/beckers_barley_wrapped_facet.html

import altair as alt
from vega_datasets import data

source = data.barley.url

alt.Chart(source).mark_point().encode(
    alt.X('median(yield):Q', scale=alt.Scale(zero=False)),
    y='variety:O',
    color='year:N',
    facet=alt.Facet('site:O', columns=2),
).properties(
    width=200,
    height=100,
)



# https://altair-viz.github.io/gallery/us_employment.html

import altair as alt
import pandas as pd
from vega_datasets import data

source = data.us_employment()
presidents = pd.DataFrame([
    {
        "start": "2006-01-01",
        "end": "2009-01-19",
        "president": "Bush"
    },
    {
        "start": "2009-01-20",
        "end": "2015-12-31",
        "president": "Obama"
    }
])

bars = alt.Chart(
    source,
    title="The U.S. employment crash during the Great Recession"
).mark_bar().encode(
    x=alt.X("month:T", title=""),
    y=alt.Y("nonfarm_change:Q", title="Change in non-farm employment (in thousands)"),
    color=alt.condition(
        alt.datum.nonfarm_change > 0,
        alt.value("steelblue"),
        alt.value("orange")
    )
)

rule = alt.Chart(presidents).mark_rule(
    color="black",
    strokeWidth=2
).encode(
    x='end:T'
).transform_filter(alt.datum.president == "Bush")

text = alt.Chart(presidents).mark_text(
    align='left',
    baseline='middle',
    dx=7,
    dy=-135,
    size=11
).encode(
    x='start:T',
    x2='end:T',
    text='president',
    color=alt.value('#000000')
)

(bars + rule + text).properties(width=600)



# https://altair-viz.github.io/gallery/simple_heatmap.html
import altair as alt
import numpy as np
import pandas as pd

# Compute x^2 + y^2 across a 2D grid
x, y = np.meshgrid(range(-5, 5), range(-5, 5))
z = x ** 2 + y ** 2

# Convert this grid to columnar data expected by Altair
source = pd.DataFrame({'x': x.ravel(),
                       'y': y.ravel(),
                       'z': z.ravel()})

alt.Chart(source).mark_rect().encode(
    x='x:O',
    y='y:O',
    color='z:Q'
)


from altair import (
    data_transformers, Chart, FacetChart, LayerChart, HConcatChart, VConcatChart, ConcatChart
)
from altair.utils.schemapi import Undefined

MAGIC_MARK_NAME = "_vf_mark{}"


def make_magic_mark_name(i):
    return MAGIC_MARK_NAME.format(i)


def transformed_data(chart, row_limit=None, exclude=None):
    """
    Evaluate the transform associated with a Chart and return the transformed
    data as one or more DataFrames

    :param chart: altair Chart, LayerChart, HConcatChart, VConcatChart, or ConcatChart
    :param row_limit: Maximum number of rows to return. None (default) for unlimited
    :param exclude: list of mark names to exclude.
    :return:
        - DataFrame of transformed data for Chart input
        - list of DataFrames for LayerChart, HConcatChart, VConcatChart, ConcatChart
    """

    from . import runtime, get_local_tz, get_inline_datasets_for_spec, vegalite_compilers

    if isinstance(chart, Chart):
        # Add dummy mark if None specified
        if chart.mark == Undefined:
            chart = chart.mark_point()

    # Deep copy chart so that we can rename marks without affecting caller
    chart = chart.copy(deep=True)

    # Rename leaf chart with magic names
    chart_names = name_chart(chart, 0, exclude=exclude)

    # Compile to vega and retrieve inline datasets
    with data_transformers.enable("vegafusion-inline"):
        vega_spec = vegalite_compilers.get()(chart.to_json(validate=False))
        inline_datasets = get_inline_datasets_for_spec(vega_spec)

    # Build mapping from mark names to vega datasets
    facet_mapping = get_facet_mapping(vega_spec)
    dataset_mapping = get_datasets_for_named_marks(vega_spec, chart_names, facet_mapping)

    # Build a list of vega dataset names that corresponds to the order
    # of the chart components
    dataset_names = []
    for chart_name in chart_names:
        if chart_name in dataset_mapping:
            dataset_names.append(dataset_mapping[chart_name])
        else:
            raise ValueError("Failed to locate all datasets")

    # Extract transformed datasets
    datasets, warnings = runtime.pre_transform_datasets(
        vega_spec,
        dataset_names,
        get_local_tz(),
        row_limit=row_limit,
        inline_datasets=inline_datasets
    )

    if isinstance(chart, (Chart, FacetChart)):
        # Return DataFrame (or None if it was excluded) if input was a simple Chart
        if not datasets:
            return None
        else:
            return datasets[0]
    else:
        # Otherwise return the list of DataFrames
        return datasets


def name_chart(chart, i=0, exclude=None):
    """
    Rename charts (and subcharts) with unique names that we can look up
    later in the compiled Vega spec

    :param chart: Chart instance
    :param i: starting name index
    :return:
    """
    exclude = exclude or []
    if isinstance(chart, (Chart, FacetChart)):
        # Perform shallow copy of chart so that we can change
        # the name
        if chart.name not in exclude:
            if chart.name in (None, Undefined):
                name = make_magic_mark_name(i)
                chart.name = name
            return [chart.name]
        else:
            return []
    else:
        if isinstance(chart, LayerChart):
            subcharts = chart.layer
        elif isinstance(chart, HConcatChart):
            subcharts = chart.hconcat
        elif isinstance(chart, VConcatChart):
            subcharts = chart.vconcat
        elif isinstance(chart, ConcatChart):
            subcharts = chart.concat
        else:
            raise ValueError(
                "transformed_data accepts an instance of "
                "Chart, FacetChart, LayerChart, HConcatChart, VConcatChart, or ConcatChart\n"
                f"Received value of type: {type(chart)}"
            )

        chart_names = []
        for subchart in subcharts:
            for name in name_chart(subchart, i=i + len(chart_names), exclude=exclude):
                chart_names.append(name)
        return chart_names


def get_facet_mapping(vega_spec, scope=()):
    facet_mapping = {}
    group_index = 0
    mark_group = get_mark_group_for_scope(vega_spec, scope) or {}
    for mark in mark_group.get("marks", []):
        if mark.get("type", None) == "group":
            # Get facet for this group
            group_scope = scope + (group_index,)
            facet = mark.get("from", {}).get("facet", None)
            if facet is not None:
                facet_name = facet.get("name", None)
                facet_data = facet.get("data", None)
                if facet_name is not None and facet_data is not None:
                    scoped_facet_data = get_scoped_data_reference(vega_spec, facet_data, scope)
                    if scoped_facet_data is not None:
                        facet_mapping[(facet_name, group_scope)] = scoped_facet_data

            # Handle children recursively
            child_mapping = get_facet_mapping(vega_spec, scope=group_scope)
            facet_mapping.update(child_mapping)
            group_index += 1

    return facet_mapping


def get_datasets_for_named_marks(vega_spec, magic_names, facet_mapping, scope=()):
    datasets = {}
    group_index = 0
    mark_group = get_mark_group_for_scope(vega_spec, scope) or {}
    for mark in mark_group.get("marks", []):
        for magic_name in magic_names:
            if mark.get("name", "") == f"{magic_name}_cell":
                data_name = mark.get("from", {}).get("facet", None).get("data", None)
                scoped_data_name = (data_name, scope)
                datasets[magic_name] = get_from_facet_mapping(scoped_data_name, facet_mapping)
                break

        name = mark.get("name", "")
        if mark.get("type", "") == "group":
            group_data_names = get_datasets_for_named_marks(
                vega_spec, magic_names, facet_mapping, scope=scope + (group_index,)
            )
            for k, v in group_data_names.items():
                datasets.setdefault(k, v)
            group_index += 1
        else:
            for magic_name in magic_names:
                if name.startswith(magic_name) and name.endswith("_marks"):
                    data_name = mark.get("from", {}).get("data", None)
                    scoped_data_name = get_scoped_data_reference(vega_spec, data_name, scope)
                    if scoped_data_name is not None:
                        datasets[magic_name] = get_from_facet_mapping(scoped_data_name, facet_mapping)
                        break

    return datasets


def get_from_facet_mapping(data_name, facet_mapping):
    while data_name in facet_mapping:
        data_name = facet_mapping[data_name]
    return data_name


def get_mark_group_for_scope(vega_spec, scope):
    group = vega_spec

    # Find group at scope
    for scope_value in scope:
        group_index = 0
        child_group = None
        for mark in group.get("marks", []):
            if mark.get("type") == "group":
                if group_index == scope_value:
                    child_group = mark
                    break
                group_index += 1
        if child_group is None:
            return None
        group = child_group

    return group


def get_datasets_for_scope(vega_spec, scope):
    group = get_mark_group_for_scope(vega_spec, scope) or {}

    # get datasets from group
    datasets = []
    for dataset in group.get("data", []):
        datasets.append(dataset["name"])

    # Add facet dataset
    facet_dataset = group.get("from", {}).get("facet", {}).get("name", None)
    if facet_dataset:
        datasets.append(facet_dataset)
    return datasets


def get_scoped_data_reference(vega_spec, data_name, usage_scope):
    for i in reversed(range(len(usage_scope) + 1)):
        scope = usage_scope[:i]
        datasets = get_datasets_for_scope(vega_spec, scope) or []
        if data_name in datasets:
            return (data_name, scope)
    return None



import altair as alt
import vegafusion as vf


def test_mime_enabler_context_manager():
    alt.data_transformers.enable("json")
    alt.renderers.enable("mimetype")

    assert alt.data_transformers.active == "json"
    assert alt.renderers.active == "mimetype"

    with vf.disable():
        assert alt.data_transformers.active == "default"
        assert alt.renderers.active == "default"

    assert alt.data_transformers.active == "json"
    assert alt.renderers.active == "mimetype"

    ctx = vf.disable()
    assert alt.data_transformers.active == "default"
    assert alt.renderers.active == "default"
    assert repr(ctx) == "vegafusion.disable()"

    with vf.enable():
        assert alt.data_transformers.active == "vegafusion-inline"
        assert alt.renderers.active == "vegafusion-mime"

    assert alt.data_transformers.active == "default"
    assert alt.renderers.active == "default"

    ctx = vf.enable()
    assert alt.data_transformers.active == "vegafusion-inline"
    assert alt.renderers.active == "vegafusion-mime"
    assert repr(ctx) == "vegafusion.enable(mimetype='html', row_limit=10000, embed_options=None)"



# https://altair-viz.github.io/gallery/table_bubble_plot_github.html

import altair as alt
from vega_datasets import data

source = "https://raw.githubusercontent.com/vega/vega-datasets/v1.31.1/data/github.csv"

alt.Chart(source).mark_circle().encode(
    x='hours(time):O',
    y='day(time):O',
    size='sum(count):Q'
)



# https://altair-viz.github.io/gallery/wilkinson-dot-plot.html

import altair as alt
import pandas as pd

source = pd.DataFrame(
    {"data":[1,1,1,1,1,1,1,1,1,1,
             2,2,2,
             3,3,
             4,4,4,4,4,4]
     }
)

alt.Chart(source).mark_circle(opacity=1).transform_window(
    id='rank()',
    groupby=['data']
).encode(
    alt.X('data:O'),
    alt.Y('id:O',
          axis=None,
          sort='descending')
).properties(height=100)



# https://altair-viz.github.io/gallery/comet_chart.html

import altair as alt
import vega_datasets

(
    alt.Chart(vega_datasets.data.barley.url)
        .transform_pivot("year", value="yield", groupby=["variety", "site"])
        .transform_fold(["1931", "1932"], as_=["year", "yield"])
        .transform_calculate(calculate="datum['1932'] - datum['1931']", as_="delta")
        .mark_trail()
        .encode(
        x=alt.X('year:O', title=None),
        y=alt.Y('variety:N', title='Variety'),
        size=alt.Size('yield:Q', scale=alt.Scale(range=[0, 12]), legend=alt.Legend(values=[20, 60], title='Barley Yield (bushels/acre)')),
        color=alt.Color('delta:Q', scale=alt.Scale(domainMid=0), legend=alt.Legend(title='Yield Delta (%)')),
        tooltip=alt.Tooltip(['year:O', 'yield:Q']),
        column=alt.Column('site:N', title='Site')

    )
        .configure_view(stroke=None)
        .configure_legend(orient='bottom', direction='horizontal')
        .properties(title='Barley Yield comparison between 1932 and 1931')
)



# https://altair-viz.github.io/gallery/world_projections.html

import altair as alt
from vega_datasets import data

source = alt.topo_feature(data.world_110m.url, 'countries')

base = alt.Chart(source).mark_geoshape(
    fill='#666666',
    stroke='white'
).properties(
    width=300,
    height=180
)

projections = ['equirectangular', 'mercator', 'orthographic', 'gnomonic']
charts = [base.project(proj).properties(title=proj)
          for proj in projections]

alt.concat(*charts, columns=2)



# https://altair-viz.github.io/gallery/window_rank.html

import altair as alt
import pandas as pd

source = pd.DataFrame(
    [
        {"team": "Germany", "matchday": 1, "point": 0, "diff": -1},
        {"team": "Germany", "matchday": 2, "point": 3, "diff": 0},
        {"team": "Germany", "matchday": 3, "point": 3, "diff": -2},
        {"team": "Mexico", "matchday": 1, "point": 3, "diff": 1},
        {"team": "Mexico", "matchday": 2, "point": 6, "diff": 2},
        {"team": "Mexico", "matchday": 3, "point": 6, "diff": -1},
        {"team": "South Korea", "matchday": 1, "point": 0, "diff": -1},
        {"team": "South Korea", "matchday": 2, "point": 0, "diff": -2},
        {"team": "South Korea", "matchday": 3, "point": 3, "diff": 0},
        {"team": "Sweden", "matchday": 1, "point": 3, "diff": 1},
        {"team": "Sweden", "matchday": 2, "point": 3, "diff": 0},
        {"team": "Sweden", "matchday": 3, "point": 6, "diff": 3},
    ]
)

color_scale = alt.Scale(
    domain=["Germany", "Mexico", "South Korea", "Sweden"],
    range=["#000000", "#127153", "#C91A3C", "#0C71AB"],
)

alt.Chart(source).mark_line().encode(
    x="matchday:O", y="rank:O", color=alt.Color("team:N", scale=color_scale)
).transform_window(
    rank="rank()",
    sort=[
        alt.SortField("point", order="descending"),
        alt.SortField("diff", order="descending"),
    ],
    groupby=["matchday"],
).properties(title="World Cup 2018: Group F Rankings")



# https://altair-viz.github.io/gallery/line_percent.html

import altair as alt
from vega_datasets import data

source = data.jobs.url

alt.Chart(source).mark_line().encode(
    alt.X('year:O'),
    alt.Y('perc:Q', axis=alt.Axis(format='%')),
    color='sex:N'
).transform_filter(
    alt.datum.job == 'Welder'
)



# https://altair-viz.github.io/gallery/airport_connections.html

import altair as alt
from vega_datasets import data

# Since these data are each more than 5,000 rows we'll import from the URLs
airports = data.airports.url
flights_airport = data.flights_airport.url

states = alt.topo_feature(data.us_10m.url, feature="states")

# Create mouseover selection
select_city = alt.selection_single(
    on="mouseover", nearest=True, fields=["origin"], empty="none"
)

# Define which attributes to lookup from airports.csv
lookup_data = alt.LookupData(
    airports, key="iata", fields=["state", "latitude", "longitude"]
)

background = alt.Chart(states).mark_geoshape(
    fill="lightgray",
    stroke="white"
).properties(
    width=600,
    height=400
).project("albersUsa")

connections = alt.Chart(flights_airport).mark_rule(opacity=0.35).encode(
    latitude="latitude:Q",
    longitude="longitude:Q",
    latitude2="lat2:Q",
    longitude2="lon2:Q"
).transform_lookup(
    lookup="origin",
    from_=lookup_data
).transform_lookup(
    lookup="destination",
    from_=lookup_data,
    as_=["state", "lat2", "lon2"]
).transform_filter(
    select_city
)

points = alt.Chart(flights_airport).mark_circle().encode(
    latitude="latitude:Q",
    longitude="longitude:Q",
    size=alt.Size("routes:Q", scale=alt.Scale(range=[0, 1000]), legend=None),
    order=alt.Order("routes:Q", sort="descending"),
    tooltip=["origin:N", "routes:Q"]
).transform_aggregate(
    routes="count()",
    groupby=["origin"]
).transform_lookup(
    lookup="origin",
    from_=lookup_data
).transform_filter(
    (alt.datum.state != "PR") & (alt.datum.state != "VI")
).add_selection(
    select_city
)

(background + connections + points).configure_view(stroke=None)



# VegaFusion
# Copyright (C) 2022, Jon Mease
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import pandas as pd
import pyarrow as pa
from decimal import Decimal
from vegafusion.transformer import to_arrow_table


def test_to_arrow_expands_categoricals():
    # Build DataFrame with one categorical column
    df = pd.DataFrame({
        "a": [1, 2, 3],
        "b": ["One", "One", "Two"]
    })
    df["b"] = df["b"].astype("category")
    assert isinstance(df["b"].dtype, pd.CategoricalDtype)

    # Convert to pyarrow table
    pa_table = to_arrow_table(df)

    # Check that pyarrow type is String (not Dictionary)
    b_type = pa_table.column("b").type
    assert b_type == pa.string()


def test_to_table_converts_decimals():
    # Build DataFrame with one Decimal column
    df = pd.DataFrame({
        "a": [1, 2, 3],
        "b": [Decimal("3.12"), Decimal("4.9"), Decimal("6")]
    })
    assert df["b"].dtype.kind == "O"

    # Convert to pyarrow table
    pa_table = to_arrow_table(df)

    # Check that pyarrow type is float64 (not Decimal128)
    b_type = pa_table.column("b").type
    assert b_type == pa.float64()


def test_to_table_with_mixed_string_int_column():
    # Build DataFrame with one Decimal column
    df = pd.DataFrame({
        "a": [1, 2, 3],
        "b": ["A", "B", 3]
    })
    assert df["b"].dtype.kind == "O"

    # Convert to pyarrow table
    pa_table = to_arrow_table(df)

    # Check that pyarrow type is float64 (not Decimal128)
    b_type = pa_table.column("b").type
    assert b_type == pa.string()


def test_to_table_with_all_conversions():
    df = pd.DataFrame({
        "a": [1, 2, 3],
        "b": ["One", "One", "Two"],
        "c": [Decimal("3.12"), Decimal("4.9"), Decimal("6")],
        "d": ["A", "B", 3]
    })
    df["b"] = df["b"].astype("category")

    # Check initial pandas column types
    assert isinstance(df["b"].dtype, pd.CategoricalDtype)
    assert df["c"].dtype.kind == "O"
    assert df["d"].dtype.kind == "O"

    # Convert to pyarrow table
    pa_table = to_arrow_table(df)

    # Check pyarrow types
    assert pa_table.column("b").type == pa.string()
    assert pa_table.column("c").type == pa.float64()
    assert pa_table.column("d").type == pa.string()



# https://altair-viz.github.io/gallery/anscombe_plot.html

import altair as alt
from vega_datasets import data

source = data.anscombe()

alt.Chart(source).mark_circle().encode(
    alt.X('X', scale=alt.Scale(zero=False)),
    alt.Y('Y', scale=alt.Scale(zero=False)),
    alt.Facet('Series', columns=2),
).properties(
    width=180,
    height=180,
)



# https://altair-viz.github.io/gallery/trellis_histogram.html
# Modification: Set height=100 to keep the screenshot from extending outside the browser window

import altair as alt
from vega_datasets import data

source = data.cars()

alt.Chart(source).mark_bar().encode(
    alt.X("Horsepower:Q", bin=True),
    y='count()',
    row='Origin'
).properties(height=100)



import altair as alt
from vega_datasets import data

source = data.population.url

select_year = alt.selection_point(
    name="Year",
    fields=["year"],
    bind=alt.binding_range(min=1900, max=2000, step=10, name="Year"),
    value={"year": 2000},
)

alt.Chart(source).mark_bar().encode(
    x=alt.X("sex:N", axis=alt.Axis(labels=False, title=None, ticks=False)),
    y=alt.Y("people:Q", scale=alt.Scale(domain=(0, 12000000)), title="Population"),
    color=alt.Color(
        "sex:N",
        scale=alt.Scale(domain=("Male", "Female"), range=["steelblue", "salmon"]),
        title="Sex",
    ),
    column=alt.Column("age:O", title="Age"),
).properties(
    width=20,
    title="U.S. Population by Age and Sex"
).add_params(
    select_year
).transform_calculate(
    "sex", alt.expr.if_(alt.datum.sex == 1, "Male", "Female")
).transform_filter(
    select_year
).configure_facet(
    spacing=8
)


class RendererTransformerEnabler:
    """
    Context manager for enabling a renderer and data
    transformer together
    """
    def __init__(self, renderer_ctx, data_transformer_ctx, repr_str=None):
        self.renderer_ctx = renderer_ctx
        self.data_transformer_ctx = data_transformer_ctx
        self.repr_str = repr_str

    def __enter__(self):
        self.renderer_ctx.__enter__()
        self.data_transformer_ctx.__enter__()
        return self

    def __exit__(self, *args, **kwargs):
        self.renderer_ctx.__exit__(*args, **kwargs)
        self.data_transformer_ctx.__exit__(*args, **kwargs)

    def __repr__(self):
        if self.repr_str:
            return self.repr_str
        else:
            return "vegafusion.utils.RendererTransformerEnabler"



# https://altair-viz.github.io/gallery/trellis_scatter_plot.html
# With reduced height

import altair as alt
from vega_datasets import data

source = data.cars()

alt.Chart(source).mark_point().encode(
    x='Horsepower:Q',
    y='Miles_per_Gallon:Q',
    row='Origin:N'
).properties(height=150)



# https://altair-viz.github.io/gallery/histogram_with_a_global_mean_overlay.html

import altair as alt
from vega_datasets import data

source = data.movies.url

base = alt.Chart(source)

bar = base.mark_bar().encode(
    x=alt.X('IMDB_Rating:Q', bin=True, axis=None),
    y='count()'
)

rule = base.mark_rule(color='red').encode(
    x='mean(IMDB_Rating):Q',
    size=alt.value(5)
)

bar + rule


import altair as alt
from vega_datasets import data

source = data.cars()

# Configure the options common to all layers
brush = alt.selection_interval()
base = alt.Chart(source).add_params(brush)

# Configure the points
points = base.mark_point().encode(
    x=alt.X('Miles_per_Gallon', title=''),
    y=alt.Y('Horsepower', title=''),
    color=alt.condition(brush, 'Origin', alt.value('grey'))
)

# Configure the ticks
tick_axis = alt.Axis(labels=False, domain=False, ticks=False)

x_ticks = base.mark_tick().encode(
    alt.X('Miles_per_Gallon', axis=tick_axis),
    alt.Y('Origin', title='', axis=tick_axis),
    color=alt.condition(brush, 'Origin', alt.value('lightgrey'))
)

y_ticks = base.mark_tick().encode(
    alt.X('Origin', title='', axis=tick_axis),
    alt.Y('Horsepower', axis=tick_axis),
    color=alt.condition(brush, 'Origin', alt.value('lightgrey'))
)

# Build the chart
y_ticks | (points & x_ticks)